#include "ChecksumFunc.h"

#include "SectList.h"
#include "PeStruct.h"
#include "PrStruct.h"
#include "Crc32.h"

static LIST_ENTRY ChecksumList;

PPEB PsGetProcessPeb(IN PEPROCESS Process);
PVOID PsGetProcessSectionBaseAddress(IN PEPROCESS Process);

VOID ProcessSections(PLIST_ENTRY Sections, PEPROCESS Pep, PVOID BaseAddr, ULONG ImageSize, UINT32 Pid, PBOOLEAN Error);
BOOLEAN IsX64Process(pimage_nt_headers32 NtHeader);

PVOID AllocVaPtr(PEPROCESS Pep, PVOID VA, ULONG VaSize, PMDL * OutMdl);
VOID FreeVaPtr(PMDL pMdl, PVOID VA);

PEPROCESS GetProcessPep(UINT32 Pid);
PPEB GetProcessPeb(PEPROCESS Pep);
BOOLEAN ProcessModules(PLIST_ENTRY Sections, UINT32 Pid);

/* Функция инициализации вышеописанной переменной */
VOID ChecksumInit()
{
    InitList(&ChecksumList);
}

/* Очистка её же */
VOID ChecksumFinalize()
{
    CleanList(&ChecksumList);
}

/* Функция-прослойка, возвращающая результат проверки целостности вызывающего процесса */
BOOLEAN CheckUserProcess()
{
    return ProcessModules(&ChecksumList, (UINT32) PsGetCurrentProcessId());
}

VOID ProcessSections(PLIST_ENTRY Sections, PEPROCESS Pep, PVOID BaseAddr, ULONG ImageSize, UINT32 Pid, PBOOLEAN Error)
{
    /* Указатель на память процесса, которая будет спроецированна в ядро */
    PVOID mPtr;
    UINT32 i, FirstSection, IsMonitored, CRC32;
    /* http://msdn.microsoft.com/en-us/library/windows/hardware/ff565421%28v=vs.85%29.aspx */
    PMDL pMdl = NULL;
    BOOLEAN IsX64 = FALSE;
    
    /* Структуры для разбора PE-файлов, взятые из библиотеки dx'a, но слегка адаптированные под ядро */
    /* http://code.google.com/p/portable-executable-library/ */
    pimage_dos_header DosHeader;
    pimage_nt_headers32 NtHeader32;
    pimage_nt_headers64 NtHeader64;
    pimage_file_header FileHeader;
    pimage_section_header SectionHeader;
    pSectionList SectList;
    
    *Error = FALSE;
    
    /* Проецируем память процесса в ядро */
    mPtr = AllocVaPtr(Pep, BaseAddr, ImageSize, &pMdl);
    if(mPtr == NULL)
    {
        CKSMLOG("(%d) Can't map VA", __LINE__);
        return;
    }
    
    /* Получаем указатель на DOS-заголовок модуля и проверяем разрядность процесса */
    DosHeader = (pimage_dos_header)mPtr;
    IsX64 = IsX64Process( (pimage_nt_headers32) ((UINT8 *)DosHeader + DosHeader->e_lfanew) );
    
    /* Получаем указатель на FileHeader в зависимости от разрядности */
    if(IsX64)
    {
        NtHeader64 = (pimage_nt_headers64) ((UINT8 *)DosHeader + DosHeader->e_lfanew);
        FileHeader = (pimage_file_header)&NtHeader64->FileHeader;
    }
    else
    {
        NtHeader32 = (pimage_nt_headers32) ((UINT8 *)DosHeader + DosHeader->e_lfanew);
        FileHeader = (pimage_file_header)&NtHeader32->FileHeader;
    }
    
    /* Получаем указатель на первую секцию */
    FirstSection = DosHeader->e_lfanew + FileHeader->SizeOfOptionalHeader + sizeof(image_file_header) + sizeof(UINT32);
    
    /* Проходим по всем загруженным секциям модуля */
    for(i = 0; i < FileHeader->NumberOfSections; i++)
    {
        SectionHeader = (pimage_section_header)((ULONG_PTR)mPtr + FirstSection + (i * sizeof(image_section_header)));
        /* Нас интересуют только секции с флагами Read, Execute, Contains code, Not Discardable */
        IsMonitored =
            (SectionHeader->Characteristics & image_scn_mem_execute)
            &&
            (SectionHeader->Characteristics & image_scn_mem_read)
            &&
            (SectionHeader->Characteristics & image_scn_cnt_code)
            &&
            !(SectionHeader->Characteristics & image_scn_mem_discardable)
            ;
        
        
        if(IsMonitored)
        {
            /* Вычисляем контрольную сумму секции и проверяем, нет ли записи соответствующей нашей секции в двусвязаном списке */
            CRC32 = Crc32Buf((UINT8 *)((ULONG_PTR)mPtr + SectionHeader->VirtualAddress), SectionHeader->Misc.VirtualSize);
            SectList = FindEntry(Sections, Pid, BaseAddr, SectionHeader->VirtualAddress, SectionHeader->Name);
            
            /* Если такой записи нет, то добавим секцию в список */
            if(SectList == NULL)
            {
                CKSMLOG("(%d) adding entry (PID=%u; BaseAddr=%08X; RVA=%08X; Name=%.8s; CRC32=%08X)", __LINE__, Pid, BaseAddr, SectionHeader->VirtualAddress, SectionHeader->Name, CRC32);
                AddEntry(Sections, Pid, BaseAddr, SectionHeader->VirtualAddress, SectionHeader->Name, CRC32);
            }
            /* Если запись есть, то сверим CRC и пометим секцию, как присутствующую в памяти */
            else if(SectList->CRC32 == CRC32)
            {
                CKSMLOG("(%d) checksum validated (PID=%u; BaseAddr=%08X; RVA=%08X; Name=%.8s; CRC32=%08X)", __LINE__, SectList->Pid, SectList->BaseAddr, SectList->RVA, SectList->Name, SectList->CRC32);
                SectList->IsPresent = 1;
            }
            /* В противном случае пометим секцию, но сообщим об ошибке проверки целостности */
            else
            {
                CKSMLOG("(%d) erroneous checksum (PID=%u; BaseAddr=%08X; RVA=%08X; Name=%.8s; CRC32=%08X~%08X)", __LINE__, SectList->Pid, SectList->BaseAddr, SectList->RVA, SectList->Name, SectList->CRC32, CRC32);
                SectList->IsPresent = 1;
                *Error = TRUE;
            }
        }
    }
    
    /* Освободим указатель на user-mode память */
    FreeVaPtr(pMdl, BaseAddr);
}

BOOLEAN IsX64Process(pimage_nt_headers32 NtHeader)
{
    return NtHeader->OptionalHeader.Magic == image_nt_optional_hdr64_magic ? TRUE : FALSE;
}

/* Проецируем юзермодный виртуальный адрес в ядро */
PVOID AllocVaPtr(PEPROCESS Pep, PVOID VA, ULONG VaSize, PMDL * OutMdl)
{
    KAPC_STATE kAPC;
    PMDL pMdl;
    PVOID Ptr = NULL;
    
    CKSMLOG("(%d) AllocVaPtr", __LINE__);
    
    /* "Подключаем" текущий поток к адресному пространству пользовательского процесса */
    KeStackAttachProcess(Pep, &kAPC);
    
    /* Получаем указатель на список физических страниц для заданного диапазона виртуальных адресов */
    pMdl = IoAllocateMdl(VA, VaSize, FALSE, FALSE, NULL);
    if(pMdl == NULL)
    {
        CKSMLOG("(%d) IoAllocateMdl", __LINE__);
        
        KeUnstackDetachProcess(&kAPC);
        return NULL;
    }
    
    /* Закрепляем страницы в памяти, а то вдруг их кто-нибудь, например, в своп скинуть попытается */
    __try
    {
        MmProbeAndLockPages(pMdl, UserMode, IoReadAccess);
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        CKSMLOG("(%d) MmProbeAndLockPages", __LINE__);

        IoFreeMdl(pMdl);
        KeUnstackDetachProcess(&kAPC);
        return NULL;
    }

    /* Проецируем физические страницы в память ядра по произвольному виртуальному адресу */
    __try
    {
        Ptr = MmMapLockedPagesSpecifyCache(pMdl, UserMode, MmCached, NULL, FALSE, NormalPagePriority);
    }
    __except(EXCEPTION_EXECUTE_HANDLER)
    {
        CKSMLOG("(%d) MmMapLockedPagesSpecifyCache", __LINE__);

        MmUnlockPages(pMdl);
        IoFreeMdl(pMdl);
        KeUnstackDetachProcess(&kAPC);
        return NULL;
    }

    if(Ptr == NULL)
    {
        CKSMLOG("(%d) MmMapLockedPagesSpecifyCache (NULL)", __LINE__);

        MmUnlockPages(pMdl);
        IoFreeMdl(pMdl);
        KeUnstackDetachProcess(&kAPC);
        return NULL;
    }
    
    /* Отсоединяем текущий поток от пользовательского процесса */
    KeUnstackDetachProcess(&kAPC);

    *OutMdl = pMdl;

    return Ptr;
}

/* Освобождаем спроецированную память и дескриптор физических страниц */
VOID FreeVaPtr(PMDL pMdl, PVOID VA)
{
    CKSMLOG("(%d) FreeVaPtr", __LINE__);
    
    MmUnmapLockedPages(VA, pMdl);
    MmUnlockPages(pMdl);
    IoFreeMdl(pMdl);
}

PEPROCESS GetProcessPep(UINT32 Pid)
{
    PEPROCESS Pep = NULL;
    NTSTATUS Status = PsLookupProcessByProcessId((HANDLE) Pid, &Pep);

    if(!NT_SUCCESS(Status))
        CKSMLOG("(%d) PsLookupProcessByProcessId (PID=%u; Status=%08X)", __LINE__, Pid, Status);

    return Pep;
}

PPEB GetProcessPeb(PEPROCESS Pep)
{
    /* Недокументированная (в MSDN) функция, которая просто возвращает элемент из структуры PEPROCESS */
    PPEB Peb = PsGetProcessPeb(Pep);
    /* MSDN не рекомендует использование этой функции, но пусть будет */
    if(!MmIsAddressValid(Peb))
    {
        CKSMLOG("(%d) MmIsAddressValid (Addr=%p)", __LINE__, Peb);
        return NULL;
    }
    
    return Peb;
}

BOOLEAN ProcessModules(PLIST_ENTRY Sections, UINT32 Pid)
{
    PPEB Peb;
    PEPROCESS Pep;
    PLDR_DATA_TABLE_ENTRY LdrEntry;
    PLIST_ENTRY Entry;
    BOOLEAN Status = FALSE, Error;
    
    /* Получаем указатель на структуру EPROCESS для заданного PID */
    Pep = GetProcessPep(Pid);
    if(Pep == NULL)
        return Status;
    
    /* Из EPROCESS получаем PEB */
    /* (http://en.wikipedia.org/wiki/Process_Environment_Block) */
    Peb = GetProcessPeb(Pep);
    if(Peb == NULL)
        return Status;
    
    CKSMLOG("(%d) PEB: %p", __LINE__, Peb);
    
    /* Проходим по списку загруженных в память модулей */
    for(Entry = Peb->Ldr->InMemoryOrderModuleList.Flink; &Peb->Ldr->InMemoryOrderModuleList != Entry; Entry = Entry->Flink)
    {
        LdrEntry = CONTAINING_RECORD(Entry, LDR_DATA_TABLE_ENTRY, InMemoryOrderLinks);
        /* Обрабатываем секции для каждого именованного модуля */
        if(LdrEntry->FullDllName.Buffer != NULL)
        {
            CKSMLOG("(%d) %p - %u - %wZ", __LINE__, LdrEntry->DllBase, LdrEntry->SizeOfImage, &LdrEntry->FullDllName);  
            ProcessSections(Sections, Pep, LdrEntry->DllBase, LdrEntry->SizeOfImage, Pid, &Error);
            if(Error)
                Status = TRUE;
        }
    }

    /* Удалим из списка все секции, связанные с текущим Pid, которые не были обработаны в цикле выше */
    DeleteUnusedEntries(Sections, Pid);
    /* Сбросим метки присутствия для секций */
    ZeroIsPresent(Sections, Pid);
    
    return Status;
}
